@@ -1,30 +1,31 @@
 /*
  *  ser2net - A program for allowing telnet connection to serial ports
- *  Copyright (C) 2001  Corey Minyard <minyard@acm.org>
+ *  Copyright (C) 2001-2020  Corey Minyard <minyard@acm.org>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  SPDX-License-Identifier: GPL-2.0-only
  *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
+ *  In addition, as a special exception, the copyright holders of
+ *  ser2net give you permission to combine ser2net with free software
+ *  programs or libraries that are released under the GNU LGPL and
+ *  with code included in the standard release of OpenSSL under the
+ *  OpenSSL license (or modified versions of such code, with unchanged
+ *  license). You may copy and distribute such a system following the
+ *  terms of the GNU GPL for ser2net and the licenses of the other code
+ *  concerned, provided that you include the source code of that
+ *  other code when and as the GNU GPL requires distribution of source
+ *  code.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *  Note that people who make modified versions of ser2net are not
+ *  obligated to grant this special exception for their modified
+ *  versions; it is their choice whether to do so. The GNU General
+ *  Public License gives permission to release a modified version
+ *  without this exception; this exception also makes it possible to
+ *  release a modified version which carries forward this exception.
  */
 
 /* This is the entry point for the ser2net program.  It reads
    parameters, initializes everything, then starts the select loop. */
 
-/* TODO
- *
- * Add some type of security
- */
-
 #include <stdio.h>
 #include <signal.h>
 #include <stdlib.h>
@@ -34,27 +35,26 @@
 #include <unistd.h>
 #include <errno.h>
 
+#include <gensio/selector.h>
+
 #include "ser2net.h"
 #include "readconfig.h"
 #include "controller.h"
-#include "utils.h"
-#include "selector.h"
 #include "dataxfer.h"
-#include "locking.h"
 #include "led.h"
 
-static char *config_file = "/etc/ser2net.conf";
-int config_port_from_cmdline = 0;
-char *config_port = NULL; /* Can be set from readconfig, too. */
+static char *config_file = SYSCONFDIR "/ser2net/ser2net.yaml";
+static bool config_file_set = false;
+static char *old_config_file = SYSCONFDIR "/ser2net.conf";
+bool admin_port_from_cmdline = false;
+char *admin_port = NULL; /* Can be set from readconfig, too. */
 static char *pid_file = NULL;
 static int detach = 1;
 int ser2net_debug = 0;
 int ser2net_debug_level = 0;
 volatile int in_shutdown = 0;
-#ifdef USE_UUCP_LOCKING
-int uucp_locking_enabled = 1;
-#endif
 #ifdef USE_PTHREADS
+#include <pthread.h>
 int num_threads = 1;
 struct thread_info {
     pthread_t id;
@@ -64,77 +64,257 @@ struct thread_info *threads;
 
 
 struct selector_s *ser2net_sel;
+struct gensio_os_funcs *so;
 char *rfc2217_signature = "ser2net";
 
 static char *help_string =
 "%s: Valid parameters are:\n"
-"  -c <config file> - use a config file besides /etc/ser2net.conf\n"
+"  -c <config file> - use a config file besides /etc/ser2net/ser2net.yaml\n"
 "  -C <config line> - Handle a single configuration line.  This may be\n"
 "     specified multiple times for multiple lines.  This is just like a\n"
 "     line in the config file.  This disables the default config file,\n"
 "     you must specify a -c after the last -C to have it read a config\n"
-"     file, too.\n"
+"     file, too.  The config file must not be yaml.\n"
 "  -p <controller port> - Start a controller session on the given TCP port\n"
 "  -P <file> - set location of pid file\n"
 "  -n - Don't detach from the controlling terminal\n"
 "  -d - Don't detach and send debug I/O to standard output\n"
-"  -l - Increate the debugging level\n"
-#ifdef USE_UUCP_LOCKING
+"  -l - Increase the debugging level\n"
 "  -u - Disable UUCP locking\n"
-#endif
 #ifdef USE_PTHREADS
 "  -t <num threads> - Use the given number of threads, default 1\n"
 #endif
 "  -b - unused (was Do CISCO IOS baud-rate negotiation, instead of RFC2217)\n"
 "  -v - print the program's version and exit\n"
-"  -s - specify a default signature for RFC2217 protocol\n";
+"  -s - specify a default signature for RFC2217 protocol\n"
+"  -Y - Handle a yaml configuration string.  This may be specified multiple\n"
+"       times; these strings are strung together as if they were one input\n"
+"       string.  This disables the defalt config file, you must specify -c\n"
+"       after the last -Y.  The config file will be processed first, if it\n"
+"       is specified, then the -Y strings in order, as if they are one\n"
+"       continguous file.  '#' characters outside of quotes will be converted\n"
+"       to newlines to make things easier to handle.  Each -Y will be\n"
+"       terminated with a newline automatically.\n";
 
-static void
-reread_config_file(void)
+static char **config_lines;
+static unsigned int num_config_lines;
+
+static bool
+str_endswith(const char *str, const char *end)
 {
-    if (config_file) {
-	char *prev_config_port = config_port;
-	config_port = NULL;
-	syslog(LOG_INFO, "Got SIGHUP, re-reading configuration");
-	readconfig_init();
-	readconfig(config_file);
-	if (config_port_from_cmdline) {
-	    /* Never override the config port from the command line. */
-	    free(config_port);
-	    config_port = prev_config_port;
-	    goto config_port_unchanged;
-	}
-	if (config_port && prev_config_port
-	    && (strcmp(config_port, prev_config_port) == 0)) {
-	    free(prev_config_port);
-	    goto config_port_unchanged;
-	}
+    unsigned int slen = strlen(str);
+    unsigned int elen = strlen(end);
 
-	if (prev_config_port) {
-	    controller_shutdown();
-	    free(prev_config_port);
-	}
+    if (elen > slen)
+	return false;
+    if (strcmp(end, str + slen - elen) == 0)
+	return true;
+    return false;
+}
 
-	if (config_port) {
-	    int rv = controller_init(config_port);
-	    if (rv == CONTROLLER_INVALID_TCP_SPEC)
-		syslog(LOG_ERR, "Invalid control port specified: %s",
-		       config_port);
-	    else if (rv == CONTROLLER_OUT_OF_MEMORY)
-		syslog(LOG_ERR, "Out of memory opening control port: %s",
-		       config_port);
-	    else if (rv == CONTROLLER_CANT_OPEN_PORT)
-		syslog(LOG_ERR, "Can't open control port: %s",
-		       config_port);
-	    if (rv) {
-		syslog(LOG_ERR, "Control port is disabled");
-		free(config_port);
-		config_port = NULL;
-	    }
-	}
+static int
+syslog_eout(struct absout *e, const char *str, ...)
+{
+    va_list ap;
+    char buf[1024];
+
+    va_start(ap, str);
+    vsnprintf(buf, sizeof(buf), str, ap);
+    va_end(ap);
+    syslog(LOG_ERR, "%s", buf);
+    return 0;
+}
+struct absout syslog_absout = { .out = syslog_eout };
+
+static int
+stderr_eout(struct absout *e, const char *str, ...)
+{
+    va_list ap;
+    char buf[1024];
+
+    va_start(ap, str);
+    vsnprintf(buf, sizeof(buf), str, ap);
+    va_end(ap);
+    syslog(LOG_ERR, "%s", buf);
+    fprintf(stderr, "%s\n", buf);
+    return 0;
+}
+static struct absout stderr_absout = { .out = stderr_eout };
+
+static int
+gensio_log_level_to_syslog(int gloglevel)
+{
+    switch (gloglevel) {
+    case GENSIO_LOG_FATAL:
+	return LOG_EMERG;
+    case GENSIO_LOG_ERR:
+	return LOG_ERR;
+    case GENSIO_LOG_WARNING:
+	return LOG_WARNING;
+    case GENSIO_LOG_INFO:
+	return LOG_INFO;
     }
- config_port_unchanged:
-    return;
+    return LOG_ERR;
+}
+
+void
+do_gensio_log(const char *name, struct gensio_loginfo *i)
+{
+    char buf[256];
+
+    vsnprintf(buf, sizeof(buf), i->str, i->args);
+    syslog(gensio_log_level_to_syslog(i->level), "%s: %s", name, buf);
+}
+
+static FILE *
+fopen_config_file(bool *is_yaml)
+{
+    FILE *instream = fopen(config_file, "r");
+
+    if (!instream) {
+	if (config_file_set) {
+	    syslog(LOG_ERR, "Unable to open config file '%s': %m",
+		   config_file);
+	    return NULL;
+	}
+	instream = fopen(old_config_file, "r");
+	if (!instream) {
+	    syslog(LOG_ERR, "Unable to open config file '%s' or old one"
+		   " '%s': %m", config_file, old_config_file);
+	    return NULL;
+	}
+	*is_yaml = false;
+    } else {
+	*is_yaml = str_endswith(config_file, ".yaml");
+    }
+
+    if (!*is_yaml) {
+	char in_buf[100];
+
+	in_buf[0] = '\0';
+	if (fgets(in_buf, sizeof(in_buf) - 1, instream) != NULL)
+	    *is_yaml = strncmp(in_buf, "%YAML", 5) == 0;
+	fseek(instream, 0, SEEK_SET);
+    }
+    return instream;
+}
+
+int
+reread_config_file(const char *reqtype, struct absout *eout)
+{
+    int rv = ENOENT;
+
+    if (config_file) {
+	FILE *instream = NULL;
+	bool is_yaml;
+
+	syslog(LOG_INFO, "Got %s, re-reading configuration", reqtype);
+	readconfig_init();
+
+	instream = fopen_config_file(&is_yaml);
+	if (!instream)
+	    goto out;
+
+	if (!admin_port_from_cmdline)
+	    controller_shutdown();
+	if (is_yaml)
+	    rv = yaml_readconfig(instream, config_lines, num_config_lines,
+				 eout);
+	else
+	    rv = readconfig(instream);
+	fclose(instream);
+
+	if (!rv)
+	    apply_new_ports(&syslog_absout);
+    }
+ out:
+    return rv;
+}
+
+void
+write_ignore_fail(int fd, const char *data, size_t count)
+{
+    ssize_t written;
+
+    while ((written = write(fd, data, count)) > 0) {
+	data += written;
+	count -= written;
+    }
+}
+
+void
+add_usec_to_time(gensio_time *tv, int usec)
+{
+#ifdef gensio_version_major
+    tv->nsecs += usec * 1000;
+    while (tv->nsecs >= 1000000000) {
+	tv->nsecs -= 1000000000;
+	tv->secs += 1;
+    }
+#else
+    tv->tv_usec += usec;
+    while (tv->tv_usec >= 1000000) {
+	tv->tv_usec -= 1000000;
+	tv->tv_sec += 1;
+    }
+#endif
+}
+
+int
+sub_time(gensio_time *left, gensio_time *right)
+{
+    gensio_time dest;
+
+#ifdef gensio_version_major
+    dest.secs = left->secs - right->secs;
+    dest.nsecs = left->nsecs - right->nsecs;
+    while (dest.nsecs < 0) {
+	dest.nsecs += 1000000000;
+	dest.secs--;
+    }
+
+    return (dest.secs * 1000000) + (dest.nsecs + 500) / 1000;
+#else
+    dest.tv_sec = left->tv_sec - right->tv_sec;
+    dest.tv_usec = left->tv_usec - right->tv_usec;
+    while (dest.tv_usec < 0) {
+	dest.tv_usec += 1000000;
+	dest.tv_sec--;
+    }
+
+    return (dest.tv_sec * 1000000) + dest.tv_usec;
+#endif
+}
+
+/* Scan for a positive integer, and return it.  Return -1 if the
+   integer was invalid. */
+int
+scan_int(const char *str)
+{
+    int rv = 0;
+
+    if (*str == '\0') {
+	return -1;
+    }
+
+    for (;;) {
+	switch (*str) {
+	case '0': case '1': case '2': case '3': case '4':
+	case '5': case '6': case '7': case '8': case '9':
+	    rv = (rv * 10) + ((*str) - '0');
+	    break;
+
+	case '\0':
+	    return rv;
+
+	default:
+	    return -1;
+	}
+
+	str++;
+    }
+
+    return rv;
 }
 
 static void
@@ -193,11 +373,12 @@ sigint_handler(int sig)
     sig_wake_selector();
 }
 
-#if USE_PTHREADS
-DEFINE_LOCK_INIT(static, config_lock)
-static int in_config_read = 0;
+static struct gensio_lock *config_lock;
 
-DEFINE_LOCK_INIT(static, maint_lock)
+static struct gensio_lock *maint_lock;
+
+#ifdef USE_PTHREADS
+static int in_config_read = 0;
 
 int ser2net_wake_sig = SIGUSR1;
 void (*finish_shutdown)(void);
@@ -205,13 +386,13 @@ void (*finish_shutdown)(void);
 void
 start_maint_op(void)
 {
-    LOCK(maint_lock);
+    so->lock(maint_lock);
 }
 
 void
 end_maint_op(void)
 {
-    UNLOCK(maint_lock);
+    so->unlock(maint_lock);
 }
 
 static void *
@@ -219,11 +400,11 @@ config_reread_thread(void *dummy)
 {
     pthread_detach(pthread_self());
     start_maint_op();
-    reread_config_file();
+    reread_config_file("SIGHUP", &syslog_absout);
     end_maint_op();
-    LOCK(config_lock);
+    so->lock(config_lock);
     in_config_read = 0;
-    UNLOCK(config_lock);
+    so->unlock(config_lock);
     return NULL;
 }
 
@@ -233,22 +414,22 @@ thread_reread_config_file(void)
     int rv;
     pthread_t thread;
 
-    LOCK(config_lock);
+    so->lock(config_lock);
     if (in_config_read) {
-	UNLOCK(config_lock);
+	so->unlock(config_lock);
 	return;
     }
     in_config_read = 1;
-    UNLOCK(config_lock);
+    so->unlock(config_lock);
 
     rv = pthread_create(&thread, NULL, config_reread_thread, NULL);
     if (rv) {
 	syslog(LOG_ERR,
 	       "Unable to start thread to reread config file: %s",
 	       strerror(rv));
-	LOCK(config_lock);
+	so->lock(config_lock);
 	in_config_read = 0;
-	UNLOCK(config_lock);
+	so->unlock(config_lock);
     }
 }
 
@@ -275,7 +456,7 @@ op_loop(void *dummy)
 	sel_select(ser2net_sel, wake_thread_send_sig, (long) &self, NULL, NULL);
 
     /* Join the threads only in the first thread.  You cannot join the
-       first thread.  Finish thw shutdown in the first thread. */
+       first thread.  Finish the shutdown in the first thread. */
     if (self == threads[0].id) {
 	int i;
 
@@ -413,7 +594,7 @@ finish_shutdown_cleanly(void)
 	sel_select(ser2net_sel, NULL, 0, NULL, &tv);
     } while(1);
 
-    sol_shutdown(); /* Free's the selector. */
+    shutdown_dataxfer();
 
     free_longstrs();
     free_tracefiles();
@@ -422,6 +603,11 @@ finish_shutdown_cleanly(void)
     if (pid_file)
 	unlink(pid_file);
 
+    if (admin_port)
+	free(admin_port);
+
+    so->free_funcs(so);
+
     exit(1);
 }
 
@@ -442,10 +628,12 @@ sig_fd_read_handler(int fd, void *cb_data)
 	shutdown_cleanly();
 
     if (reread_config && !in_shutdown) {
-#if USE_PTHREADS
+#ifdef USE_PTHREADS
 	thread_reread_config_file();
 #else
-	reread_config_file();
+	start_maint_op();
+	reread_config_file("SIGHUP", &syslog_absout);
+	end_maint_op();
 #endif
     }
 }
@@ -485,9 +673,11 @@ setup_signals(void)
     if (err)
 	goto out;
 
-    sel_set_fd_handlers(ser2net_sel, sig_fd_watch, NULL, sig_fd_read_handler,
-			NULL, NULL, NULL);
-    sel_set_fd_read_handler(ser2net_sel, sig_fd_watch, SEL_FD_HANDLER_ENABLED);
+    err = sel_set_fd_handlers(ser2net_sel, sig_fd_watch, NULL,
+			      sig_fd_read_handler, NULL, NULL, NULL);
+    if (!err)
+	sel_set_fd_read_handler(ser2net_sel, sig_fd_watch,
+				SEL_FD_HANDLER_ENABLED);
 
  out:
     if (err) {
@@ -496,6 +686,23 @@ setup_signals(void)
     }
 }
 
+static void
+ser2net_gensio_logger(struct gensio_os_funcs *o, enum gensio_log_levels level,
+		      const char *log, va_list args)
+{
+    int priority = LOG_INFO;
+
+    switch (level) {
+    case GENSIO_LOG_FATAL:   priority = LOG_CRIT; break;
+    case GENSIO_LOG_ERR:     priority = LOG_ERR; break;
+    case GENSIO_LOG_WARNING: priority = LOG_WARNING; break;
+    case GENSIO_LOG_INFO:    priority = LOG_INFO; break;
+    case GENSIO_LOG_DEBUG:   priority = LOG_DEBUG; break;
+    }
+
+    vsyslog(priority, log, args);
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -504,8 +711,11 @@ main(int argc, char *argv[])
 #ifdef USE_PTHREADS
     char *end;
 #endif
-    char **config_lines;
-    int num_config_lines = 0;
+    int print_when_ready = 0;
+    FILE *instream = NULL;
+    enum { CONFIG_NONE, CONFIG_YAML, CONFIG_OLD } config_type = CONFIG_NONE;
+
+    gensio_set_progname("ser2net");
 
     config_lines = malloc(sizeof(*config_lines));
     if (!config_lines) {
@@ -526,6 +736,10 @@ main(int argc, char *argv[])
 	}
 
 	switch (argv[i][1]) {
+	case 'r':
+	    print_when_ready = 1;
+	    break;
+
 	case 'n':
 	    detach = 0;
 	    break;
@@ -537,16 +751,34 @@ main(int argc, char *argv[])
 
 	case 'l':
 	    ser2net_debug_level++;
+	    gensio_set_log_mask(gensio_get_log_mask() |
+				1 << (ser2net_debug_level + 2));
 	    break;
 
 	case 'b':
 	    break;
 
 	case 'C':
+	case 'Y':
+	    if (argv[i][1] == 'C') {
+		if (config_type == CONFIG_YAML) {
+		mixed_config_lines:
+		    fprintf(stderr,
+			    "You cannot mix yaml and old config lines.\n");
+		    arg_error(argv[0]);
+		}
+		config_type = CONFIG_OLD;
+	    } else {
+		if (config_type == CONFIG_OLD)
+		    goto mixed_config_lines;
+		config_type = CONFIG_YAML;
+	    }
+
 	    /* Get a config line. */
 	    i++;
 	    if (i == argc) {
-		fprintf(stderr, "No config line specified with -C\n");
+		fprintf(stderr, "No config line specified with -%c\n",
+			argv[i][1]);
 		arg_error(argv[0]);
 	    }
 	    num_config_lines++;
@@ -558,6 +790,7 @@ main(int argc, char *argv[])
 	    }
 	    config_lines[num_config_lines - 1] = argv[i];
 	    config_file = NULL;
+	    config_file_set = true;
 	    break;
 
 	case 'c':
@@ -568,6 +801,7 @@ main(int argc, char *argv[])
 		arg_error(argv[0]);
 	    }
 	    config_file = argv[i];
+	    config_file_set = true;
 	    break;
 
 	case 'p':
@@ -577,12 +811,12 @@ main(int argc, char *argv[])
 		fprintf(stderr, "No control port specified with -p\n");
 		arg_error(argv[0]);
 	    }
-	    config_port = strdup(argv[i]);
-	    if (!config_port) {
+	    admin_port = strdup(argv[i]);
+	    if (!admin_port) {
 		fprintf(stderr, "Could not allocate memory for -p\n");
 		exit(1);
 	    }
-	    config_port_from_cmdline = 1;
+	    admin_port_from_cmdline = true;
 	    break;
 
 	case 'P':
@@ -594,11 +828,9 @@ main(int argc, char *argv[])
 	    pid_file = argv[i];
 	    break;
 
-#ifdef USE_UUCP_LOCKING
 	case 'u':
-	    uucp_locking_enabled = 0;
+	    gensio_uucp_locking_enabled = 0;
 	    break;
-#endif
 
 	case 'v':
 	    printf("%s version %s\n", argv[0], VERSION);
@@ -651,43 +883,96 @@ main(int argc, char *argv[])
 	exit(1);
     }
 
+    so = gensio_selector_alloc(ser2net_sel, ser2net_wake_sig);
+    if (!so) {
+	fprintf(stderr, "Could not alloc ser2net gensio selector\n");
+	exit(1);
+    }
+    so->vlog = ser2net_gensio_logger;
+
+    config_lock = so->alloc_lock(so);
+    if (!config_lock) {
+	fprintf(stderr, "Could not alloc ser2net config lock\n");
+	exit(1);
+    }
+
+    maint_lock = so->alloc_lock(so);
+    if (!maint_lock) {
+	fprintf(stderr, "Could not alloc ser2net maint lock\n");
+	exit(1);
+    }
+
     setup_signals();
 
-    err = sol_init();
+    err = init_dataxfer();
     if (err) {
 	fprintf(stderr,
-		"Could not initialize IPMI SOL: '%s'\n",
-		strerror(-err));
+		"Could not initialize dataxfer: '%s'\n",
+		strerror(err));
 	exit(1);
     }
 
     if (ser2net_debug && !detach)
 	openlog("ser2net", LOG_PID | LOG_CONS | LOG_PERROR, LOG_DAEMON);
 
-    readconfig_init();
-    for (i = 0; i < num_config_lines; i++)
-	handle_config_line(config_lines[i], strlen(config_lines[i]));
-    free(config_lines);
-    if (config_file) {
-	if (readconfig(config_file) == -1)
-	    exit(1);
+    err = readconfig_init();
+    if (err) {
+	fprintf(stderr,
+		"Could not initialize defaults: '%s'\n",
+		gensio_err_to_str(err));
+	exit(1);
     }
 
-    if (config_port != NULL) {
-	int rv;
-	rv = controller_init(config_port);
-	if (rv == CONTROLLER_INVALID_TCP_SPEC) {
-	    fprintf(stderr, "Invalid control port specified: %s\n",
-		    config_port);
-	    arg_error(argv[0]);
+    if (admin_port)
+	controller_init(admin_port, NULL, NULL);
+
+    if (config_type == CONFIG_OLD) {
+	for (i = 0; i < num_config_lines; i++)
+	    handle_config_line(config_lines[i], strlen(config_lines[i]));
+    }
+    if (config_file) {
+	bool is_yaml;
+
+	if (strcmp(config_file, "-") == 0) {
+	    instream = stdin;
+	    is_yaml = true;
+	} else {
+	    instream = fopen_config_file(&is_yaml);
+	    if (!instream)
+		exit(1);
 	}
-	if (rv == CONTROLLER_CANT_OPEN_PORT) {
-	    fprintf(stderr, "Unable to open control port, see syslog: %s\n",
-		    config_port);
-	    exit(1);
+	if (is_yaml) {
+	    if (config_type == CONFIG_OLD) {
+		fprintf(stderr, "You cannot mix old style config lines "
+			"and a yaml config file.\n");
+		exit(1);
+	    }
+	    config_type = CONFIG_YAML;
+	} else {
+	    if (config_type == CONFIG_YAML) {
+		fprintf(stderr, "You cannot mix yaml config lines "
+			"and an old style config file.\n");
+		exit(1);
+	    }
+	    config_type = CONFIG_OLD;
 	}
     }
 
+    if (config_type == CONFIG_YAML || instream)	{
+	int rv;
+
+	if (config_type == CONFIG_YAML)
+	    rv = yaml_readconfig(instream, config_lines, num_config_lines,
+				 &stderr_absout);
+	else
+	    rv = readconfig(instream);
+	if (rv == -1)
+	    exit(1);
+	if (instream && instream != stdin)
+	    fclose(instream);
+    }
+    apply_new_ports(&syslog_absout);
+
     if (detach) {
 	int pid;
 
@@ -726,8 +1011,17 @@ main(int argc, char *argv[])
     make_pidfile();
 
     start_threads();
+
+    if (print_when_ready) {
+	printf("Ready\n");
+	fflush(stdout);
+    }
+
     op_loop(NULL);
 
+    so->free_lock(maint_lock);
+    so->free_lock(config_lock);
+    so->free_funcs(so);
     sel_free_selector(ser2net_sel);
 
     return 0;
